#include <Arduino.h>
#include "LedMatrix.h"
#include "Snake.h"

Snake::Snake(LedMatrix* matrix)
{
    /* init matrix */
    this->matrix = matrix;

    /* init snake */
    for (int i = 0; i < SNAKE_MAX_SIZE; i++)
    {
        this->snake[i][0] = -1; this->snake[i][1] = -1;
    }
}

void Snake::initGame()
{
    /* init fields */
    this->direction     = this->startingDirection;
    this->moveLength_ms = 500;
    this->prevMove      = millis() / this->moveLength_ms;

    /* init snake */
    for (int i = 0; i < SNAKE_START_SIZE; i++)
    {
        if (this->direction == SNAKE_DIRECTION_LEFT)
        {
            this->snake[i][0] = 3; this->snake[i][1] = 4 + i;
        }
        else if (this->direction == SNAKE_DIRECTION_RIGHT)
        {
            this->snake[i][0] = 3; this->snake[i][1] = 3 - i;
        }
        else if (this->direction == SNAKE_DIRECTION_UP)
        {
            this->snake[i][0] = 4 + i; this->snake[i][1] = 3;
        }
        else if (this->direction == SNAKE_DIRECTION_DOWN)
        {
            this->snake[i][0] = 3 - i; this->snake[i][1] = 3;
        }
    }
    for (int i = SNAKE_START_SIZE; i < SNAKE_MAX_SIZE; i++)
    {
        this->snake[i][0] = -1; this->snake[i][1] = -1;
    }

    /* init food */
    this->newFood();

    /* init matrix */
    this->matrix->clear();
    this->matrix->turnOn(this->snake[0][0], this->snake[0][1], this->snakeColor);
    this->matrix->turnOn(this->snake[1][0], this->snake[1][1], this->snakeColor);
    this->matrix->turnOn(this->snake[2][0], this->snake[2][1], this->snakeColor);
}

void Snake::loop()
{
    /* set up matrix content if snake playing */
    if (this->gameStatus == SNAKE_STATUS_PLAYING)
    {
        /* update position */
        int currentMove = millis() / moveLength_ms;
        if (this->prevMove != currentMove)
        {
            this->moveSnake();
            this->prevMove = currentMove;
        }

        /* show food */
        this->matrix->turnOn(this->food[0], this->food[1], this->foodColor);

        /* ensure whole snake is turned on */
        int size = this->getSize();
        for (int i = 0; i < size; i++) {
            this->matrix->turnOn(this->snake[i][0], this->snake[i][1], this->snakeColor);
        }
    }

    /* update display */
    this->matrix->display();
}

void Snake::setSnakeColor(int color)
{
    this->snakeColor = color;
}

void Snake::setFoodColor(int color)
{
    this->foodColor = color;
}

void Snake::restart()
{
    this->matrix->clear();
    this->initGame();
    this->gameStatus = SNAKE_STATUS_PLAYING;
}

void Snake::changeDirection(int direction)
{
    /* check this is a valid direction */
    if (direction == SNAKE_DIRECTION_LEFT || direction == SNAKE_DIRECTION_RIGHT
        || direction == SNAKE_DIRECTION_UP || direction == SNAKE_DIRECTION_DOWN)
    {
        /* if game is started, change the current direction of the snake */
        if (this->gameStatus == SNAKE_STATUS_PLAYING)
        {
            bool isOppositeDirection =
                   (direction == SNAKE_DIRECTION_LEFT && this->direction == SNAKE_DIRECTION_RIGHT)
                || (direction == SNAKE_DIRECTION_RIGHT && this->direction == SNAKE_DIRECTION_LEFT)
                || (direction == SNAKE_DIRECTION_UP && this->direction == SNAKE_DIRECTION_DOWN)
                || (direction == SNAKE_DIRECTION_DOWN && this->direction == SNAKE_DIRECTION_UP);

            /* if the new direction is the opposite direction, it is not updated as
            the snake will instantly be in its queue */
            if (!isOppositeDirection)
            {
                this->direction = direction;
            }
        }
        /* if game is not yet started, change the direction of the snake on start */
        else
        {
            this->startingDirection = direction;
        }
    }

}

bool Snake::isInWall()
{
    for (int i = 1; this->isSnake(i); i++)
    {
        if (abs(this->snake[i-1][0] - this->snake[i][0]) > 1
            || abs(this->snake[i-1][1] - this->snake[i][1]) > 1)
        {
            return true;
        }
    }
    return false;
}

bool Snake::isInQueue()
{
    /* test if one part of the snake body is on another part of the snake body */
    for (int i = 0; isSnake(i); i++)
    {
        int* currentPosition = this->snake[i];
        for (int j = i + 1; isSnake(j); j++)
        {
            if (currentPosition[0] == this->snake[j][0]
                && currentPosition[1] == this->snake[j][1])
            {
                return true;
            }
        }
    }
}

bool Snake::isEating()
{
    return this->snake[0][0] == this->food[0]
        && this->snake[0][1] == this->food[1];
}

void Snake::grow()
{
    int queuePos;
    for (queuePos = 0; isSnake(queuePos); queuePos++);
    this->snake[queuePos][0] = -2;
    this->snake[queuePos][1] = -2;
}

void Snake::newFood()
{
    this->matrix->set(this->food[0], this->food[1], LED_OFF);

    bool positionIsAlreadyUsedBySnake = true;
    while (positionIsAlreadyUsedBySnake)
    {
        this->food[0] = random(0, MATRIX_SIZE);
        this->food[1] = random(0, MATRIX_SIZE);

        positionIsAlreadyUsedBySnake = false;
        for (int i = 0; isSnake(i); i++)
        {
            if (this->snake[i][0] == this->food[0]
                && this->snake[i][1] == this->food[1])
            {
                positionIsAlreadyUsedBySnake = true;
                break;
            }
        }
    }
}

void Snake::accelerate()
{
    this->moveLength_ms -= this->moveLength_ms / 10;
    this->prevMove      = millis() / this->moveLength_ms;
}

void Snake::slowDown()
{
    this->moveLength_ms += this->moveLength_ms / 9;
    this->prevMove      = millis() / this->moveLength_ms;
}

int Snake::getSize()
{
    int size;
    for (size = 0; isSnake(size); size++);
    return size + 1;
}

void Snake::loose()
{
    if (this->gameStatus == SNAKE_STATUS_PLAYING) {
        this->gameStatus = SNAKE_STATUS_LOOSE;
        this->matrix->clear();
    }
}

void Snake::win()
{
    if (this->gameStatus == SNAKE_STATUS_PLAYING) {
        this->gameStatus = SNAKE_STATUS_WIN;
        this->matrix->clear();
    }
}

bool Snake::isSnake(int position)
{
    return position >= 0 && position < SNAKE_MAX_SIZE
            && this->snake[position][0] >= 0 && this->snake[position][1] >= 0;
}

void Snake::moveSnake()
{
    int size = this->getSize() - 1;

    /* grow the snake or remove the queue to move forward */
    if (size < SNAKE_MAX_SIZE
        && (this->snake[size][0] == -2 || this->snake[size][1] == -2))
    {
        /* the snake grows */
        this->snake[size][0] = this->snake[size - 1][0];
        this->snake[size][1] = this->snake[size - 1][1];
    }
    else
    {
        /* remove queue */
        this->matrix->turnOff(this->snake[size - 1][0], this->snake[size - 1][1]);
    }


    /* shift head -> body -> queue */
    /* starts at [size] to allow growing using -2 */
    for (int i = size - 1; i > 0; i--)
    {
        this->snake[i][0] = this->snake[i - 1][0];
        this->snake[i][1] = this->snake[i - 1][1];
    }

    /* add new head */
    switch (this->direction)
    {
        case SNAKE_DIRECTION_RIGHT:
            this->snake[0][1] = (this->snake[0][1] + 1) % MATRIX_SIZE;
            break;
        case SNAKE_DIRECTION_LEFT:
            this->snake[0][1]--;
            if (this->snake[0][1] < 0) { this->snake[0][1] = MATRIX_SIZE - 1; }
            break;
        case SNAKE_DIRECTION_UP:
            this->snake[0][0]--;
            if (this->snake[0][0] < 0) { this->snake[0][0] = MATRIX_SIZE - 1; }
            break;
        case SNAKE_DIRECTION_DOWN:
            this->snake[0][0] = (this->snake[0][0] + 1) % MATRIX_SIZE;
            break;
    }
}
